import asyncio
from io import BufferedReader
import logging
from typing import Optional

from telegram.ext import ApplicationBuilder, MessageHandler, ChatMemberHandler
from telegram.error import TelegramError
from telegram import (InputFile, InputMediaAudio, InputMediaDocument,
                      InputMediaPhoto, InputMediaVideo, Update)

from wiring.bot_base import Bot
from wiring.errors.bot_api_error import BotApiError
from wiring.logging_options import DEFAULT_LOGGING_OPTIONS
from wiring.platforms.telegram._entities_converter import telegram_entities_converter


class TelegramBot(Bot):
    def __init__(self, token: str, logging_options=DEFAULT_LOGGING_OPTIONS):
        super().__init__('telegram')
        self.client = ApplicationBuilder().token(token).build()

        self.__setup_event_handlers()

        self.logger = logging.getLogger('telegram')

        if logging_options['handler']:
            self.logger.addHandler(logging_options['handler'])
            self.logger.setLevel(logging_options['level'])

    def __setup_event_handlers(self):
        # mess, need to move it to a different module/class
        async def handle_update(update: Update, context):
            if update.message is not None:
                multi_platform_chat = (
                    telegram_entities_converter
                    .convert_to_multi_platform_chat_group(update.message.chat)
                )

                multi_platform_message = (
                    telegram_entities_converter
                    .convert_to_multi_platform_message(update.message)
                )

                self._run_event_handlers('message', multi_platform_message)

                self._check_message_for_command(multi_platform_message)

                for new_member in update.message.new_chat_members or []:
    
                    self._run_event_handlers(
                        'join',
                        telegram_entities_converter.convert_to_multi_platform_user(
                            new_member,
                            multi_platform_chat
                        )
                    )

                if update.message.left_chat_member is not None:
                    self._run_event_handlers(
                        'leave',
                        telegram_entities_converter.convert_to_multi_platform_user(
                            update.message.left_chat_member,
                            multi_platform_chat
                        )
                    )

        # registering the same handler for each needed update
        # because i cant find global update handler solution
        self.client.add_handler(ChatMemberHandler(callback=handle_update))
        self.client.add_handler(MessageHandler(filters=None, callback=handle_update))

    async def start(self):
        await self.client.initialize()

        if self.client.updater is None:
            raise Exception('cant start polling in telegram bot.'
                            + '\'client.updater\' attribute is \'None\'')

        try:
            self.event_listening_coroutine = asyncio.create_task(
                self.client.updater.start_polling()
            )
            await self.client.start()
        except Exception:
            await self.client.stop()

    async def stop(self):
        await self.client.stop()
        await self.client.shutdown()

    async def send_message(self, chat_id, text: str,
                           reply_message_id=None,
                           files: Optional[list[BufferedReader]] = None):
        try:
            if files is not None:
                await self.client.bot.send_media_group(
                    chat_id,
                    [self.__convert_stream_to_telegram_media(file) for file in files],
                    caption=text,
                    reply_to_message_id=reply_message_id
                )
                return

            await self.client.bot.send_message(chat_id, text,
                                               reply_to_message_id=reply_message_id)
        except TelegramError as telegram_error:
            raise BotApiError('telegram', telegram_error.message)

    async def get_chats_from_group(self, chat_group_id: int):
        try:
            return [
                telegram_entities_converter.convert_to_multi_platform_chat(
                    await self.client.bot.get_chat(chat_group_id)
                )
            ]
        except TelegramError as telegram_error:
            raise BotApiError('telegram', telegram_error.message)

    def __convert_stream_to_telegram_media(self, stream: BufferedReader):
        file = InputFile(stream)
        mimetype = file.mimetype

        if mimetype.startswith('video'):
            return InputMediaVideo(media=file.input_file_content)
        elif mimetype.startswith('image'):
            return InputMediaPhoto(media=file.input_file_content)
        elif mimetype.startswith('audio'):
            return InputMediaAudio(media=file.input_file_content)
        else:
            return InputMediaDocument(media=file.input_file_content)



from abc import ABC, abstractmethod

from wiring.multi_platform_resources import (MultiPlatformChatGroup,
                                               MultiPlatformMessage,
                                               MultiPlatformChat,
                                               MultiPlatformUser)


class ToMultiPlatformConverter(ABC):
    @abstractmethod
    def convert_to_multi_platform_message(self, message) -> MultiPlatformMessage:
        pass

    @abstractmethod
    def convert_to_multi_platform_user(self, user) -> MultiPlatformUser:
        pass

    @abstractmethod
    def convert_to_multi_platform_chat_group(self, chat_group) -> MultiPlatformChatGroup:
        pass

    @abstractmethod
    def convert_to_multi_platform_chat(self, chat) -> MultiPlatformChat:
        pass



from abc import ABC, abstractmethod
import asyncio
from dataclasses import dataclass
from io import BufferedReader
from typing import Any, Callable, Coroutine, Literal, Optional, Awaitable

from wiring.multi_platform_resources import MultiPlatformMessage, MultiPlatformChat


CommandHandler = Callable[[Any, MultiPlatformMessage, list[str]], Coroutine]

Event = Literal['message', 'join', 'leave']


@dataclass
class Command:
    name: list[str] | str
    handler: CommandHandler


@dataclass
class EventHandler:
    event: Event
    do_on_event: Callable[[Any, Any], Coroutine]


class Bot(ABC):
    def __init__(self,
                 platform: Optional[str] = None):
        self.platform = platform
        self.commands_prefix = '/'
        self.commands = []

        self.event_listening_coroutine: Optional[Awaitable] = None

        self._event_handlers: list[EventHandler] = []

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass

    @abstractmethod
    async def send_message(self, chat_id,
                           text: str,
                           reply_message_id: Any = None,
                           files: Optional[list[BufferedReader]] = None):
        """sends message in the chat

        Args:
            files (list): (optional) images streams to read and embed as a files.
                **closes the streams automatically after reading**

        Raises:
            NotMessageableChatError: if message cant be sent in target chat
            BotApiError: if error occurred on some platform api interaction
            ValueError: the files list is not of the appropriate size
        """
        pass

    @abstractmethod
    async def get_chats_from_group(self, chat_group_id) -> list[MultiPlatformChat]:
        """fetches a group of connected chats, for example, a discord server

        Returns:
            list of chats from specified group. if target platform doesnt support
            chat groups, they are considered identical to chats,
            so it returns list of one chat converted from request chat group

        Raises:
            BotApiError: if error occurred on some platform api interaction
        """
        pass

    async def setup_commands(self, commands: list[Command], prefix: str = '/'):
        self.commands_prefix = prefix
        self.commands = commands

    def add_event_handler(self, event: Event, handler: Callable[[Any, Any], Coroutine]):
        """
        adds a handler function that will be called when specified event occurs
            with `Bot` object and event data as arguments

        supported events:
            - `message` - when a message sent in any chat bot are member of.
                ignores current bot's messages to prevent recursion
            - `join` - when someone joins in chat/chat group bot are member of
            - `leave` - when someone leaves the chat/chat group bot are member of
        """
        self._event_handlers.append(EventHandler(event, handler))

    async def __aenter__(self):
        await self.start()

    async def __aexit__(self, *args):
        await self.stop()

    def _run_event_handlers(self, event: Event, event_data=None):
        [asyncio.create_task(handler.do_on_event(self, event_data))
         for handler in self._event_handlers
         if handler.event == event]

    def _check_message_for_command(self, message: MultiPlatformMessage):
        def has_command(text: str, command: Command):
            cleaned_text = text.removeprefix(self.commands_prefix).strip().casefold()

            if isinstance(command.name, list):
                return len([name for name in command.name
                            if cleaned_text == name.casefold()]) > 0
            else:
                return command.name.casefold() == cleaned_text

        if not message.text or not message.text.startswith(self.commands_prefix):
            return

        message_parts = message.text.split(' ')

        if len(message_parts) == 0:
            return

        matched_commands = [
            some_command for some_command in self.commands
            if has_command(message_parts[0], some_command)
        ]

        if len(matched_commands) > 0:
            asyncio.create_task(
                matched_commands[0].handler(self, message, message_parts[1:])
            )



import discord

from wiring.multi_platform_resources import (MultiPlatformChatGroup, MultiPlatformMessage,
                                      MultiPlatformChat, MultiPlatformUser)
from wiring.platforms.discord.channels import MessageableChannel
from wiring._to_multi_platform_converter import ToMultiPlatformConverter


class DiscordEntitiesConverter(ToMultiPlatformConverter):
    def convert_to_multi_platform_chat_group(self, chat_group: discord.Guild):
        return MultiPlatformChatGroup('discord', chat_group.id, chat_group.name)

    def convert_to_multi_platform_chat(
        self, chat: MessageableChannel
    ):
        name = None

        if (not isinstance(chat, discord.PartialMessageable)
                and not isinstance(chat, discord.DMChannel)):
            name = chat.name

        return MultiPlatformChat('discord',
                                 chat.id,
                                 name)

    def convert_to_multi_platform_user(self, user: discord.User | discord.Member):
        if isinstance(user, discord.Member):
            return MultiPlatformUser('discord', user.name,
                                     self.convert_to_multi_platform_chat_group(user.guild))

        return MultiPlatformUser('discord', user.name, None)

    def convert_to_multi_platform_message(
        self, message: discord.Message
    ):
        chat = None

        if message.guild is not None:
            chat = self.convert_to_multi_platform_chat_group(message.guild)

        return MultiPlatformMessage(
            'discord',
            message.id,
            chat,
            self.convert_to_multi_platform_chat(message.channel),
            message.content,
            self.convert_to_multi_platform_user(message.author)
        )


discord_entities_converter = DiscordEntitiesConverter()



import logging
from typing import Optional

from wiring.bot_base import Bot, Command, Event
from wiring.logging_options import DEFAULT_LOGGING_OPTIONS
from wiring.multi_platform_resources import MultiPlatformValue, PlatformSpecificValue


class PlatformBotNotFoundError(Exception):
    def __init__(self, requested_platform: str):
        super().__init__(f'bot with platform \'{requested_platform}\' was not added')


class MultiPlatformBot(Bot):
    def __init__(self, logging_options=DEFAULT_LOGGING_OPTIONS):
        super().__init__()
        self.platform_bots: list[Bot] = []

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging_options['level'])

    async def start(self):
        self.logger.info('started')
        for bot in self.platform_bots:
            await bot.start()

    async def stop(self):
        for bot in self.platform_bots:
            await bot.stop()

    async def listen_to_events(self):
        for bot in self.platform_bots:
            if bot.event_listening_coroutine is not None:
                await bot.event_listening_coroutine

    async def send_message(self, chat_id: MultiPlatformValue, text: str,
                           reply_message_id: MultiPlatformValue = {},
                           files: Optional[list] = None):
        for bot in self.platform_bots:
            if bot.platform not in chat_id:
                continue

            platform_chat_id = chat_id.get(bot.platform)
            platform_reply_message_id = reply_message_id.get(bot.platform)

            if platform_chat_id is not None:
                self.logger.info(f'sending message to chat \'{platform_chat_id}\' '
                                 + f'on \'{bot.platform}\'')

                await bot.send_message(platform_chat_id, text,
                                       platform_reply_message_id,
                                       files)

    async def get_chats_from_group(self, chat_group_id: PlatformSpecificValue):
        """fetches chats grouped in some entity like discord server

        Raises:
            PlatformBotNotFoundError: if bot for specified platform was not added
        """
        needed_bots = [bot for bot in self.platform_bots
                       if bot.platform == chat_group_id['platform']]

        if len(needed_bots) == 0:
            raise PlatformBotNotFoundError(chat_group_id['platform'])

        return await needed_bots[0].get_chats_from_group(chat_group_id['value'])

    def add_event_handler(self, event: Event, handler):
        super().add_event_handler(event, handler)
        for bot in self.platform_bots:
            bot.add_event_handler(event, handler)

    async def setup_commands(self, commands: list[Command], prefix: str = '/'):
        for bot in self.platform_bots:
            await bot.setup_commands(commands, prefix)

        await super().setup_commands(commands, prefix)



from wiring.multi_platform_bot import MultiPlatformBot, PlatformBotNotFoundError
from wiring.multi_platform_resources import (PlatformSpecificValue, MultiPlatformChat,
                                               MultiPlatformChatGroup, MultiPlatformId,
                                               MultiPlatformMessage, MultiPlatformUser,
                                               MultiPlatformValue, Platform)
from wiring.bot_base import Bot, Command, CommandHandler



import discord

from typing import Union


PartialMessageableChannel = Union[discord.TextChannel, discord.VoiceChannel,
                                  discord.StageChannel, discord.Thread,
                                  discord.DMChannel, discord.PartialMessageable,
                                  discord.CategoryChannel, discord.ForumChannel]
MessageableChannel = Union[PartialMessageableChannel, discord.GroupChannel]



import logging
from typing import Optional
from telegram import Chat, Message, User, ChatFullInfo

from wiring._to_multi_platform_converter import ToMultiPlatformConverter
from wiring.multi_platform_resources import (MultiPlatformChatGroup, MultiPlatformMessage,
                                      MultiPlatformChat, MultiPlatformUser)


class TelegramEntitiesConverter(ToMultiPlatformConverter):
    def convert_to_multi_platform_chat_group(self, chat_group: Chat):
        logger = logging.getLogger('telegram')
        logger.warning('chat group is the same entity as the chat for telegram')

        return MultiPlatformChatGroup('telegram',
                                      chat_group.id,
                                      chat_group.title or chat_group.full_name)

    def convert_to_multi_platform_chat(self, chat: Chat | ChatFullInfo):
        return MultiPlatformChat('telegram',
                                 chat.id,
                                 chat.title or chat.full_name)

    def convert_to_multi_platform_user(self, user: User,
                                       from_chat: Optional[MultiPlatformChatGroup] = None):
        return MultiPlatformUser('telegram', user.username or user.full_name, from_chat)

    def convert_to_multi_platform_message(
        self, message: Message
    ):
        chat = self.convert_to_multi_platform_chat_group(message.chat)

        author_user = (self.convert_to_multi_platform_user(message.from_user, chat)
                       if message.from_user is not None else None)

        return MultiPlatformMessage(
            'telegram',
            message.id,
            chat,
            self.convert_to_multi_platform_chat(message.chat),
            message.text,
            author_user
        )


telegram_entities_converter = TelegramEntitiesConverter()



import logging
from typing import Optional, TypedDict


class LoggingOptions(TypedDict):
    handler: Optional[logging.Handler]
    level: int


DEFAULT_LOGGING_OPTIONS: LoggingOptions = {'handler': None, 'level': logging.WARNING}



from wiring.multi_platform_resources import Platform


class NotMessageableChatError(Exception):
    def __init__(self, platform: Platform, chat_id=None):
        if chat_id is not None:
            super().__init__(f'cant send message in chat \'{chat_id}\' on {platform}')
        else:
            super().__init__(f'cant send message in specified chat on {platform}')



import asyncio
from typing import Any, Callable, Optional

import discord

from wiring.bot_base import Bot
from wiring.errors.bot_api_error import BotApiError
from wiring.platforms.discord._entities_converter import discord_entities_converter
from wiring.logging_options import DEFAULT_LOGGING_OPTIONS
from wiring.errors.not_messageable_chat import NotMessageableChatError


class CustomClient(discord.Client):
    def __init__(self,
                 intents: discord.Intents,
                 event_handlers: dict[str, Callable]):
        super().__init__(intents=intents)
        self._event_handlers = event_handlers

    async def on_message(self, message: discord.Message):
        if not self.user or message.author.id == self.user.id:
            return

        self.__run_event_handler_if_exists(
            'all', 'message', discord_entities_converter.convert_to_multi_platform_message(message)
        )

        self.__run_event_handler_if_exists(
            'message', discord_entities_converter.convert_to_multi_platform_message(message)
        )

    async def on_member_join(self, member: discord.Member):
        self.__run_event_handler_if_exists(
            'all', 'join', discord_entities_converter.convert_to_multi_platform_user(member)
        )

        self.__run_event_handler_if_exists(
            'join', discord_entities_converter.convert_to_multi_platform_user(member)
        )

    async def on_member_remove(self, member: discord.Member):
        multi_platform_user = (discord_entities_converter
                               .convert_to_multi_platform_user(member))
        
        self.__run_event_handler_if_exists(
            'all', 'leave', multi_platform_user
        )

        self.__run_event_handler_if_exists(
            'leave', multi_platform_user
        )

    def __run_event_handler_if_exists(self, event: str, *args):
        do_on_event = self._event_handlers.get(event)

        if do_on_event is not None:
            do_on_event(*args)


class DiscordBot(Bot):
    def __init__(self, token: str, logging_options=DEFAULT_LOGGING_OPTIONS):
        super().__init__('discord')

        intents = discord.Intents.default()
        intents.message_content = True
        intents.members = True

        self.client = CustomClient(intents, {
            'all': lambda event, data: self._run_event_handlers(event, data),
            'message': self._check_message_for_command,
        })
        self._token = token

        discord.utils.setup_logging(handler=logging_options['handler']
                                    or discord.utils.MISSING,
                                    level=logging_options['level'])

    async def start(self):
        await self.client.login(self._token)
        self.event_listening_coroutine = asyncio.create_task(self.client.connect())

    async def stop(self):
        await self.client.close()

    async def send_message(self, chat_id: int, text: str,
                           reply_message_id: Optional[int] = None,
                           files: Optional[list] = None):
        channel: Any = await self.client.fetch_channel(chat_id)

        if not hasattr(channel, 'send'):
            raise NotMessageableChatError('discord', channel.id)

        files = [discord.File(file) for file in files or []]

        try:
            if reply_message_id is not None:
                message: discord.Message = await channel.fetch_message(reply_message_id)
                await message.reply(text, files=files)
            else:
                await channel.send(text, files=files)
        except discord.HTTPException as discord_error:
            raise BotApiError('discord', discord_error.text, discord_error.status)

    async def get_chats_from_group(self, chat_group_id: int):
        try:
            channels = await (await self.client.fetch_guild(chat_group_id)).fetch_channels()
            return [discord_entities_converter.convert_to_multi_platform_chat(channel)
                    for channel in channels]
        except discord.HTTPException as discord_http_error:
            raise BotApiError('discord', discord_http_error.text,
                              discord_http_error.status)
        except discord.errors.InvalidData:
            raise BotApiError('discord', 'received invalid data from api')



from dataclasses import dataclass
from typing import Any, Literal, Optional, TypedDict


Platform = Literal['discord', 'telegram']


MultiPlatformValue = dict[Platform, Any]


MultiPlatformId = str | int


class PlatformSpecificValue(TypedDict):
    platform: Platform
    value: Any


@dataclass
class MultiPlatformChatGroup:
    """group of chats like discord server or discord private messages

    if platform doesnt support chat groups, they are considered
    identical to chats
    """
    platform: Platform
    id: MultiPlatformId
    name: Optional[str]


@dataclass
class MultiPlatformChat:
    """
    telegram chat, discord channel or other platform chat where message was sent
    """
    platform: Platform
    id: MultiPlatformId
    name: Optional[str]


@dataclass
class MultiPlatformUser:
    platform: Platform
    username: str
    from_chat_group: Optional[MultiPlatformChatGroup]


@dataclass
class MultiPlatformMessage():
    platform: Platform
    id: MultiPlatformId
    chat_group: Optional[MultiPlatformChatGroup]
    chat: Optional[MultiPlatformChat]
    text: Optional[str]
    author_user: Optional[MultiPlatformUser]



from typing import Optional

from wiring.multi_platform_resources import Platform


class BotApiError(Exception):
    """
    common class that represents an http error occurred on some platform
    api interaction

    Attributes:
        status_code: http error status code
    """
    def __init__(self, platform: Platform, explanation: Optional[str] = None,
                 status_code: Optional[int] = None):
        """creates common http error that occurred on some platform api interaction

        Args:
            status_code: http error status code
        """

        self.explanation = explanation
        self.status_code = status_code

        error_text = f'failed to perform some action through {platform} api'

        if explanation is not None:
            error_text = explanation + f'. platform: {platform}'

        if status_code is not None:
            error_text += f', status code: {status_code}'

        super().__init__(error_text)
